{-# LANGUAGE Unsafe #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE StandaloneKindSignatures #-}
{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE ImportQualifiedPost #-}
{-# LANGUAGE FlexibleContexts #-}

{-# OPTIONS_GHC -Wno-unsafe #-}

module GHC.Generics.Coerce
(GenericCoerce(..), genericCoerce)
where

import Prelude (id, ($), (.))

import GHC.Generics (U1, (:*:) ((:*:)), (:+:) (L1, R1), M1 (M1), K1 (K1), Generic)
import GHC.Generics qualified as Generics

import Data.Coerce (coerce, Coercible)
import Data.Kind (Type, Constraint)

type GenericCoerce :: (Type -> Type) -> (Type -> Type) -> Constraint

class GenericCoerce a b where
  gfrom :: a i -> b i
  gto :: b i -> a i

-- unit type data
instance GenericCoerce U1 U1 where
  gfrom :: U1 i -> U1 i
  gfrom = id
  gto :: U1 i -> U1 i
  gto = id

-- product type combinations
instance (GenericCoerce a c, GenericCoerce b d) => GenericCoerce (a :*: b) (c :*: d) where
  gfrom :: (:*:) a b i -> (:*:) c d i
  gfrom (a :*: b) = gfrom a :*: gfrom b
  gto   :: (:*:) c d i -> (:*:) a b i
  gto   (a :*: b) = gto a :*: gto b

-- sum type options
instance (GenericCoerce a c, GenericCoerce b d) => GenericCoerce (a :+: b) (c :+: d) where
  gfrom :: (:+:) a b i -> (:+:) c d i
  gfrom (L1 a) = L1 $ gfrom a
  gfrom (R1 b) = R1 $ gfrom b
  gto :: (:+:) c d i -> (:+:) a b i
  gto (L1 a) = L1 $ gto a
  gto (R1 b) = R1 $ gto b

-- types and constructors
instance GenericCoerce a a' => GenericCoerce (M1 i_ c a) (M1 i_' c' a') where
  gfrom :: M1 i_ c a i -> M1 i_' c' a' i
  gfrom (M1 x) = M1 $ gfrom x
  gto :: M1 i_' c' a' i -> M1 i_ c a i
  gto (M1 x) = M1 $ gto x

-- constants and parameters
instance Coercible a a' => GenericCoerce (K1 i_ a) (K1 i_' a') where
  gfrom :: K1 i_ a i -> K1 i_' a' i
  gfrom (K1 x) = K1 $ coerce x
  gto :: K1 i_' a' i -> K1 i_ a i
  gto (K1 x) = K1 $ coerce x


-- convenience functions

genericCoerce :: (Generic a, Generic b, GenericCoerce (Generics.Rep a) (Generics.Rep b)) => a -> b
genericCoerce = Generics.to . gfrom . Generics.from
